/* form-submit.c -- Process form submission

	Copyright 1994,2005,2006
		by Mark E. Mallett, MV Communications, Inc.

	See the package's "LICENSE" file for terms.


   Invoker passes form information in the PATH_INFO, as:

	directory/formname:path-tag


*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "mvform.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-cgi.h>

#include "formmisc.h"


/* Local Definitions */


/* External data referenced */

extern	char	*Doc_out_mimeP;		/* Output mime type */
extern	BSTR	*GbsP;			/* General BSTR */
extern	BSTR	*GtkbsP;		/* BSTR for tokens */
extern	TKNS	*GtknsP;		/* Tokens pool */

/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

	char	*DocumentRootP;		/* Document_root */
	char	Formdir[PATH_MAX +1];	/* Dir for this form */

/* Local routines and forward references */



/* Private data */

static	int	Logfmode;		/* access mode of file */
static	char	*LogfomodeP = NULL;	/* Open mode of logfile */
static	char	*LogfnameP;
static	char	*Log_from_fnameP;
static	char	*Email_fnameP;
static	char	*Email_from_nameP;
static	char	*Email_to_nameP;
static	char	*Path_infoP;
static	char	*Path_info_nameP;
static	char	Path_info_required = FALSE;
static	char	*Path_tagP;
static	char	*Return_document_nameP;

main( argc, argv )
	int		argc;
	char		**argv;
{
	char		*tagP;
	char		*errP;
	char		*formnameP;	/* name of form */
	char		*cP, *c1P;	/* Character ptrs */
	void		*qP;
	struct stat	statblk;

    /* put stderr onto stdout */
    close( 2 );
    dup2( 1, 2 );

    if ( !formmisc_init() )
	exit( 0 );

    /* Collect arguments */
    if ( ( errP = qargs_in( (char *)NULL ) ) != NULL ) {
	report_error( errP );
	exit( 0 );
    }

    cP = getenv( "PATH_INFO" );
    if ( ( cP != NULL ) && ( *cP == '/' ) )
	++cP;			/* Skip '/' */
    if ( ( cP == NULL ) || ( *cP == NUL ) ) {
	report_error( "no form name was supplied." );
	exit( 0 );
    }

    /* The path name looks like this:

	pathname/formname:form-argument

       pathname is relative to DOCUMENT_ROOT and is required;
       formname is required;
       form-argument is any optional info for the form.
    */
    DocumentRootP = getenv( "DOCUMENT_ROOT" );
    if ( ( DocumentRootP == NULL ) || ( *DocumentRootP != '/' ) ) {
	report_error( "Can't determine DOCUMENT_ROOT" );
	exit( 0 );
    }

    if ( ( strlen( DocumentRootP ) + strlen( cP ) + 2 ) > PATH_MAX ) {
	report_error( "Buffer Overflow" );
	exit( 0 );
    }

    c1P = strrchr( cP, ':' );
    if ( c1P != NULL ) {
	*c1P++ = NUL;
	Path_infoP = c1P;
    }
    else
	Path_infoP = NULL;

    c1P = strrchr( cP, '/' );
    if ( c1P == NULL ) {
	report_error( "No directory specified for form." );
	exit( 0 );
    }
    *c1P++ = NUL;
    formnameP = c1P;
    if ( snprintf( &Formdir[0], sizeof(Formdir),
		   "%s/%s", DocumentRootP, cP ) >= sizeof(Formdir) ) {
	report_error( "forms path too large." );
	exit( 0 );
    }

    /* Get parameters for submission */
    if ( ! get_submit_defs( formnameP ) )
	exit ( 0 );

    /* If path info required, make sure it is present. */
    if ( Path_info_required && ( Path_infoP == NULL ) ) {
	report_error( "No path_info given as required by form." );
	exit( 0 );
    }

    /* From this point on we run as the owner of the form directory */
    if ( stat( &Formdir[0], &statblk ) < 0 ) {
	report_error( "Can't access named form directory." );
	exit( 0 );
    }
    if ( statblk.st_uid < 10 ) {
	report_error( "Invalid directory owner." );
	exit( 0 );
    }
    if ( setgid( statblk.st_gid ) < 0 ) {
	report_error( "Can't change to form group." );
	exit( 0 );
    }
    if ( setuid( statblk.st_uid ) < 0 ) {
	report_error( "Can't change to form owner." );
	exit( 0 );
    }

    /* If there's a log filename, log the form */
    if ( LogfnameP != NULL )
	if ( !log_form() )
	    exit( 0 );

    /* If there's an email-to name, send email */
    if ( Email_to_nameP != NULL )
	if ( !email() )
	    exit( 0 );

    /* If there's a return document, cat it */
    begin_doc_out();
    if ( Return_document_nameP != NULL )
	return_document( Return_document_nameP );
    else
	printf( "Form submitted.\n" );

    exit( 0 );
}
/*

*//* get_submit_defs( formnameP )

	Scan the submit-defs file

Accepts :

	formnameP	Name of form

Returns :

	< value >	TRUE if OK
			FALSE if not (with error already generated h)

*/

int
get_submit_defs ARGLIST( (formnameP) )
    FARG( char		*formnameP )	/* Name of form */
{
	int		status;
	int		harderrorC;
	int		submiterrorC;
	char		*strP;
	char		*argP;		/* Some arg ptr */
	void		*qP;		/* Query arg ptr */
	FILE		*fP;
	char		fdfname[PATH_MAX +1];	/* Name of formdef file */

    /* Try to open the definition file */
    if ( snprintf( &fdfname[0], sizeof(fdfname),
		   "%s/%s.%s",
		   Formdir, formnameP, SUBMITDEF ) >= sizeof(fdfname ) ) {
	report_error( "definition file path too large." );
	return ( FALSE );
    }
    fP = fopen( &fdfname[0], "r" );
    if ( fP == NULL ) {
	report_error( "Can't find definition for form \"%s\"", formnameP );
	return ( FALSE );
    }

    /* Read and parse the defs */
    for ( status = TRUE, harderrorC = 0, submiterrorC = 0;
          status && ( getformline( fP, GbsP ) >= 0 );
	     ) {
	tkline( GbsP, GtknsP );

	if ( GtknsP->tp_tC == 0 )	/* Blank line */
	    continue;

	/* Switch on first token. */
	strP = (char *)TKN_P( GtknsP, 0 );
	// if ( strcmp( strP, "#" ) == 0 )
	if ( *strP == '#' )
	    continue;			/* Comment */

	else if ( stricmp( strP, "Email-file" ) == 0 ) {
	    Email_fnameP = newstr( "get_submit_defs",
			   "email file name", (char *)TKN_P( GtknsP, 1 ) );
	}
	else if ( stricmp( strP, "Email-from" ) == 0 ) {
	    Email_from_nameP = newstr( "get_submit_defs",
			   "email-from name", (char *)TKN_P( GtknsP, 1 ) );
	}
	else if ( stricmp( strP, "Email-to" ) == 0 ) {
	    Email_to_nameP = newstr( "get_submit_defs",
			   "email-to name", (char *)TKN_P( GtknsP, 1 ) );
	}
	else if ( stricmp( strP, "Log-from" ) == 0 ) {
	    Log_from_fnameP = newstr( "get_submit_defs",
			   "log_from name", (char *)TKN_P( GtknsP, 1 ) );
	}
	else if ( stricmp( strP, "Log-into" ) == 0 ) {
	    LogfnameP = newstr( "get_submit_defs",
			   "log file name", (char *)TKN_P( GtknsP, 1 ) );
	    if ( strchr( LogfnameP, '/' ) != NULL ) {
		report_error( "'/' not allowed in logfile name." );
		return ( FALSE );
	    }

	    /* Check for permissions and other option(s) */
	    Logfmode = 0600;		/* Assume private */
	    LogfomodeP = "a";		/* Assume append open mode */
	    if ( GtknsP->tp_tC > 2 ) {
		if ( stricmp( (char *)TKN_P( GtknsP, 2 ), "Append" ) == 0 )
		    LogfomodeP = "a";
		else if ( stricmp( (char *)TKN_P( GtknsP, 2 ),
				   "Private" ) == 0 )
		    Logfmode = 0600;
		else if ( stricmp( (char *)TKN_P( GtknsP, 2 ),
				   "Public" ) == 0 )
		    Logfmode = 0644;
		else if ( ( GtknsP->tp_tC > 3 ) &&
		          ( stricmp( (char *)TKN_P( GtknsP, 2 ),
				     "Mode" ) == 0 ) ) {
		    Logfmode = strtoq( (char *)TKN_P( GtknsP, 3 ),
				       (char **)NULL, 8 );
		    if ( ( Logfmode < 0400 ) || ( Logfmode > 0777 ) ) {
			report_error( "Invalid log-into mode %o", Logfmode );
			Logfmode = 0644;
		    }
		}
		else if ( stricmp( (char *)TKN_P( GtknsP, 2 ),
				   "Replace" ) == 0 )
		    LogfomodeP = "w";
		else
		    report_error( "Unknown log-into keyword \"%s\"",
					(char *)TKN_P( GtknsP, 2 ) );
	    }
	}
	else if ( stricmp( strP, "Path-required" ) == 0 ) {
	    Path_info_required = TRUE;
	}
	else if ( stricmp( strP, "Path-tag" ) == 0 ) {
	    Path_tagP = newstr( "get_submit_defs",
			       "path tag", (char *)TKN_P( GtknsP, 1 ) );

	    /* Enter the tag and value into the environment */
	    setenv( Path_tagP, Path_infoP, TRUE );
	}
	else if ( stricmp( strP, "Return-document" ) == 0 ) {
	    Return_document_nameP = newstr( "get_submit_defs",
			   "return document name",
			   (char *)TKN_P( GtknsP, 1 ) );

#ifdef	SERVER_SEES_MIME_TYPE		/* This would be nice, eh? */
	    /* Check for document type, translate to mime type */
	    if ( GtknsP->tp_tC > 2 ) {
		argP = (char *)TKN_P( GtknsP, 2 );
		if ( stricmp( argP, "html" ) == 0 )
		    Doc_out_mimeP = "text/html";
		else if ( stricmp( argP, "php" ) == 0 )
		    Doc_out_mimeP = "text/x-httpd-php";
		else if ( stricmp( argP, "shtml" ) == 0 )
		    Doc_out_mimeP = "text/x-server-parsed-html";
		else {
		    report_error( "Unknown return-document type \"%s\"",
				  argP );
		    Doc_out_mimeP = NULL;
		}
	    }
	    else {
		/* No document type given, make sure we clear it here */
		Doc_out_mimeP = NULL;
	    }
#endif	/* SERVER_SEES_MIME_TYPE */

	}
	else if ( stricmp( strP, "Tag" ) == 0 ) {
	    /* Add a tag and a value */
	    if ( GtknsP->tp_tC != 3 ) {
	        report_error( "Error in submit-defs:<br>\n" );
		report_error( "%s<br>\n", (char *)GbsP->bs_bP );
		++harderrorC;
	    }
	    else if ( ( qP = qarg_find_tag( (char *)TKN_P( GtknsP, 1 ) ) )
				!= NULL ) {
		/* Tag already exists; replace the value */
		if ( qarg_value_set( qP, (char *)TKN_P( GtknsP, 2 ) )
				== NULL ) {
		    report_error( "Error in submit-defs: can't replace tag %s<br>\n",
				(char *)TKN_P( GtknsP, 1 ) );
		    ++harderrorC;
		}
	    }
	    else if ( !qarg_put( (char *)TKN_P( GtknsP, 1 ),
				 (char *)TKN_P( GtknsP, 2 ) ) ) {
		report_error( "Error in submit-defs: can't add tag %s<br>\n",
				(char *)TKN_P( GtknsP, 1 ) );
		++harderrorC;
	    }
	}
	else if ( stricmp( strP, "Tag-Default" ) == 0 ) {
	    /* Add a tag and a value but only if tag isn't there. */
	    if ( GtknsP->tp_tC != 3 ) {
	        report_error( "Error in submit-defs:<br>\n" );
		report_error( "%s<br>\n", (char *)GbsP->bs_bP );
		++harderrorC;
	    }
	    else if ( ( ( qP = qarg_find_tag( (char *)TKN_P( GtknsP, 1 ) ) ) == NULL ) ||
	         ( isblankstr( qarg_value( qP ) ) ) ) {
	        if ( !qarg_put( (char *)TKN_P( GtknsP, 1 ),
			        (char *)TKN_P( GtknsP, 2 ) ) )
		    ++harderrorC;
	    }
	}
	else if ( stricmp( strP, "Tag-required" ) == 0 ) {
	    if ( ( ( qP = qarg_find_tag( (char *)TKN_P( GtknsP, 1 ) ) ) == NULL ) ||
	         ( isblankstr( qarg_value( qP ) ) ) ) {
		report_error( "Error: form tag \"%s\" is missing!<br>\n",
		        (char *)TKN_P( GtknsP, 1 ) );
		++submiterrorC;
	    }
	}
	else {
	    report_error( "Bad token \"%s\" in form definition file.",
				strP );
	    status = FALSE;
	}
    }

    fclose( fP );

    if ( harderrorC > 0 ) {
        report_error( "<p>" );
	report_error( "The errors noted above indicate an internal problem,\n" );
	report_error( "preventing successfull interpretation of your submission.\n" );
	status = FALSE;
    }

    if ( submiterrorC > 0 ) {
        report_error( "<p>" );
	report_error( "When you see this page it means you've left out some\n" );
	report_error( "important data in the form.  The fields are noted\n" );
	report_error( "above.  Please go back and try again.\n" );
	status = FALSE;
    }

    return ( status );
}
/*

*//* log_form()

	Log the form information in the given form file

Accepts :

	LogfnameP	Name of log file

Returns :

	< value >	TRUE if OK
			FALSE if not

*/

int
log_form NOARGLIST
{
	int		status;
	int		rL;		/* Remaining result string length */
	int		i;
	char		*tagP, *lasttagP;
	char		*valP;
	char		*rP;		/* Result string ptr */
	FILE		*fP;		/* file ptr */
	void		*qP;		/* Query arg pointer */
	time_t		timenow;
	mode_t		oldmode;
	struct tm	*tmP;
	char		rspbuf[BIGBUFSIZE];	/* Result string */
	char		logfname[PATH_MAX +1]; /* Name of logfile */
	char		logtfname[PATH_MAX +1]; /* Name of template */

    if ( !makefilename( &logfname[0], sizeof(logfname), LogfnameP ) )
	return ( FALSE );

    oldmode = umask( 0777 - Logfmode );
    fP = fopen( &logfname[0], LogfomodeP );
    umask( oldmode );

    if ( fP == NULL ) {
	report_error( "Can't open logfile to file form data." );
	return ( FALSE );
    }
    if ( Log_from_fnameP != NULL ) {
	/* Log from a template file. */
	if ( !makefilename( &logtfname[0], sizeof(logtfname), Log_from_fnameP ) )
	    return ( FALSE );
	status = formfilecat( fP, &logtfname[0] );
	fclose( fP );
	return ( status );
    }

    /* No template, use a standard log format. */
    rP = &rspbuf[0];
    rL = sizeof(rspbuf) -1;

    /* If there's a path-tag, stick the path_info into the buffer */
    if ( Path_tagP != NULL ) {
	i = snprintf( rP, rL, "%s\t%s", Path_tagP, Path_infoP );
	if ( i > rL ) {
	    report_error( "Buffer overflow." );
	    return ( FALSE );
	}
	rP += i;
	rL -= i;
    }

    /* Go through all the tags */
    qP = qarg_first();
    lasttagP = "";
    while ( qP != NULL ) {
	tagP = qarg_tag( qP );
	valP = qarg_value( qP );

	if ( stricmp( tagP, lasttagP ) != 0 ) {
	    /* Start of new tag */
	    lasttagP = tagP;
	    if ( rP != &rspbuf[0] ) {
		if ( --rL < 0 ) {
		    report_error( "Buffer overflow" );
		    return ( FALSE );
		}
		*rP++ = '\t';
	    }
	    i = strlen( tagP );
	    if ( (i +1) > rL ) {
		report_error( "Buffer overflow" );
		return ( FALSE );
	    }
	    strcpy( rP, tagP );
	    rP += i;
	    *rP++ = '\t';
	    rL -= (i +1);
	}
	else {
	    /* Same tag, separate with comma from last value */
	    if ( --rL < 0 ) {
		report_error( "Buffer overflow" );
		return ( FALSE );
	    }
	    *rP++ = ',';
	}

	/* Add value */
	i = strlen( valP );
	if ( i > rL ) {
	    report_error( "Buffer overflow" );
	    return ( FALSE );
	}
	strcpy( rP, valP );
	rP += i;
	rL -= i;

	lasttagP = tagP;
	qP = qarg_next( qP );
    }
    *rP = NUL;

    /* File this thing */
    time( &timenow );
    tmP = localtime( &timenow );
    fprintf( fP, "%04d%02d%02d.%02d%02d%02d	%s\n",
		   tmP->tm_year + 1900, tmP->tm_mon +1, tmP->tm_mday,
		   tmP->tm_hour, tmP->tm_min, tmP->tm_sec,
		   &rspbuf[0] );
    fclose( fP );
    return ( TRUE );
}
/*

*//* email()

	Send email based on the form submission

Accepts :

	Email_to_nameP	Name to send email to

Returns :

	< value >	TRUE if OK
			FALSE if not

*/

int
email NOARGLIST
{
	int		status;
	char		*fromP;		/* "from" address, if given */
	char		*tagP, *lasttagP;
	char		*valP;
	FILE		*pP;		/* Pipe handle */
	void		*qP;
	char		cmdbuf[BIGBUFSIZE];
	char		filename[200];

    /* Get from address */
    if ( ( fromP = Email_from_nameP ) == NULL )
	fromP = "form-submit";

    /* Start up a mailer */
    if ( snprintf( &cmdbuf[0], sizeof(cmdbuf),
		   "%s -i -f %s %s",
		   SENDMAIL, fromP, Email_to_nameP ) >= sizeof(cmdbuf) ) {
	report_error( "sendmail command path too long." );
	return ( FALSE );
    }
    pP = popen( &cmdbuf[0], "w" );
    if ( pP == NULL ) {
	report_error( "Can't start mailer on server" );
	exit( 0 );
    }

    if ( Email_fnameP != NULL ) {
	/* Send the mail from a file */
	if ( ! makefilename( &filename[0], sizeof(filename), Email_fnameP ) )
	    return( FALSE );

	/* Use formfilecat to send it to the mailer */
	status = formfilecat( pP, &filename[0] );
	pclose( pP );
	return ( status );
    }

    /* No file name given: give canned message */
    fprintf( pP, "From: %s\n", fromP );
    fprintf( pP, "To: %s\n", Email_to_nameP );
    fprintf( pP, "Subject: Mail from form submission (via form-submit)\n" );
    fprintf( pP, "\n" );

    fprintf( pP, "Form Submission at MV Communications, Inc.\n" );
    fprintf( pP, "\n" );

    /* If there's a path-tag, report it. */
    if ( Path_tagP != NULL )
	fprintf( pP, "Path-Tag(%s):   %s\n\n", Path_tagP, Path_infoP );


    /* Go through all the tags. */
    qP = qarg_first();
    lasttagP = "";
    while ( qP != NULL ) {
	tagP = qarg_tag( qP );
	valP = qarg_value( qP );

	if ( stricmp( tagP, lasttagP ) != 0 ) {
	    /* Start of new tag */
	    if ( *lasttagP != NUL )
		fprintf( pP, "\n" );
	    fprintf( pP, "%s:	", tagP );
	}
	else
	    /* Same tag, separate with comma from last value */
	    fprintf( pP, ", " );

	/* Send value */
	fprintf( pP, "%s", valP );

	lasttagP = tagP;
	qP = qarg_next( qP );
    }

    if ( *lasttagP != NUL )
	fprintf( pP, "\n" );

    pclose( pP );

    return ( TRUE );
}
/*

*//* return_document( docnameP )

	Send the form to stdout

Accepts :

	docnameP		Name of document

Returns :

	< nothing >

*/

int
return_document ARGLIST( (docnameP) )
    FARG( char		*docnameP )	/* Name of document */
{
	char		filename[PATH_MAX +1]; /* Name of file */

    /* Get name of file */
    if ( !makefilename( &filename[0], sizeof(filename), docnameP ) )
	return ( FALSE );

    /* Use formfilecat to cat it to stdout */
    return ( formfilecat( stdout, &filename[0] ) );
}
